"""
All things query
"""
from math import ceil
from typing import Mapping, Iterable, Optional

from .elastic import es


def build_body(queries: Iterable[str] = None, filters: Mapping = None):
    def parse_filter(field, filter_):
        if 'value' in filter_:
            return {"term": {field: filter_['value']}}
        elif 'range' in filter_:
            return {"range": {field: filter_['range']}}

    def parse_query(q):
        return {"query_string": {"query": q}}

    def parse_queries(qs):
        if len(qs) == 1:
            return parse_query(qs[0])
        else:
            return {"bool": {"should": [parse_query(q) for q in qs]}}

    if not queries and not filters:
        return {'match_all': {}}
    fs = [parse_filter(*item) for item in filters.items()] if filters else []
    if queries:
        fs.append(parse_queries(queries))

    return {"bool": {"filter": fs}}


class QueryResult:
    def __init__(self, data, n=None, per_page=None, page=None, page_count=None, scroll_id=None):
        if n and page_count is None:
            page_count = ceil(n.get("value") / per_page)
        self.data = data
        self.total_count = n
        self.page = page
        self.page_count = page_count
        self.per_page = per_page
        self.scroll_id = scroll_id

    def as_dict(self):
        if self.scroll_id:
            meta = dict(scroll_id=self.scroll_id)
        else:
            meta = dict(total_count=self.total_count, per_page=self.per_page, page_count=self.page_count,
                        page=self.page)
        return dict(meta=meta, results=self.data)


def query_documents(index: str, queries: Iterable[str] = None, *, page: int = 0, per_page: int = 10,
                    scroll=None, scroll_id: str = None, fields: Iterable[str] = None,
                    filters: Mapping[str, Mapping] = None, **kwargs) -> Optional[QueryResult]:
    """
    Conduct a query_string query, returning the found documents
    It will return at most per_page results.
    In normal (paginated) mode, the next batch can be  requested by incrementing the page parameter.
    If the scroll parameter is given, the result will contain a scroll_id which can be used to get the next batch.
    In case there are no more documents to scroll, it will return None
    :param index: The name of the index
    :param queries: The elasticsearch query_strings
    :param page: The number of the page to request (starting from zero)
    :param per_page: The number of hits per page
    :param scroll: if not None, will create a scroll request rather than a paginated request. Parmeter should
                   specify the time the context should be kept alive, or True to get the default of 2m.
    :param scroll_id: if not None, should be a previously returned context_id to retrieve a new page of results
    :param fields: if not None, specify a list of fields to retrieve for each hit
    :param filters: if not None, a dict of filters: {field: {'value': value}} or
                    {field: {'range': {'gte/gt/lte/lt': value, 'gte/gt/..': value, ..}}
    :param kwargs: Additional elements passed to Elasticsearch.search(), for example:
           sort=col1:desc,col2
    :return: a QueryResult, or None if there is not scroll result anymore
    """
    if scroll or scroll_id:
        # set scroll to default also if scroll_id is given but no scroll time is known
        kwargs['scroll'] = '2m' if (not scroll or scroll is True) else scroll

    if scroll_id:
        result = es.scroll(scroll_id=scroll_id, **kwargs)
        if not result['hits']['hits']:
            return None
    else:
        body = build_body(queries, filters)

        if fields:
            fields = fields if isinstance(fields, list) else list(fields)
            kwargs['_source'] = fields
        if not scroll:
            kwargs['from_'] = page * per_page
        print(index,body,per_page,kwargs)
        result = es.search(index=index, body={'query': body}, size=per_page, **kwargs)


    data = [dict(_id=hit['_id'], **hit['_source']) for hit in result['hits']['hits']]
    if scroll_id:
        return QueryResult(data, scroll_id=scroll_id)
    elif scroll:
        return QueryResult(data, n=result['hits']['total'], per_page=per_page, scroll_id=result['_scroll_id'])
    else:
        return QueryResult(data, n=result['hits']['total'], per_page=per_page,  page=page)



import hashlib
import json
import logging
from typing import Mapping, List

from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

es = Elasticsearch()

SYS_INDEX = "amcat4system"
SYS_MAPPING = "sys"
REQUIRED_FIELDS = ["title", "date", "text"]
HASH_FIELDS = REQUIRED_FIELDS + ["url"]
DEFAULT_QUERY_FIELDS = HASH_FIELDS

ES_MAPPINGS = {
   'int': {"type": "long"},
   'date': {"type": "date", "format": "strict_date_optional_time"},
   'num': {"type": "double"},
   'keyword': {"type": "keyword"},
   'text': {"type": "text"},
   'object': {"type": "object"},
   }


def setup_elastic(*hosts):
    """
    Check whether we can connect with elastic
    """
    global es
    logging.debug("Connecting with elasticsearch at {}".format(hosts or "(default: localhost:9200)"))
    es = Elasticsearch(hosts or None)
    if not es.ping():
        raise Exception("Cannot connect to elasticsearch server [{}]".format(hosts))
    if not es.indices.exists(SYS_INDEX):
        logging.info("Creating amcat4 system index: {}".format(SYS_INDEX))
        es.indices.create(SYS_INDEX)


def _list_indices(exclude_system_index=True) -> [str]:
    """
    List all indices on the connected elastic cluster.
    You should probably use the methods in amcat4.index rather than this.
    """
    result = es.indices.get("*")
    return [x for x in result.keys() if not (exclude_system_index and x == SYS_INDEX)]


def _create_index(name: str) -> None:
    """
    Create a new index
    You should probably use the methods in amcat4.index rather than this.
    """
    fields = {'text': ES_MAPPINGS['text'],
              'title': ES_MAPPINGS['text'],
              'date': ES_MAPPINGS['date'],
              'url': ES_MAPPINGS['keyword']}
    body = {'mappings':
                     {'properties': fields}}
    # body = {'mappings':
    #             {'properties': fields}}
    es.indices.create(index=name, body=body)


def _delete_index(name: str, ignore_missing=False) -> None:
    """
    Delete an index
    You should probably use the methods in amcat4.index rather than this.
    :param name: The name of the new index (without prefix)
    :param ignore_missing: If True, do not throw exception if index does not exist
    """
    es.indices.delete(index=name, ignore=([404] if ignore_missing else []))


def _get_hash(document):
    """
    Get the hash for a document
    """
    hash_dict = {key: document.get(key) for key in HASH_FIELDS}
    hash_str = json.dumps(hash_dict, sort_keys=True, ensure_ascii=True).encode('ascii')
    m = hashlib.sha224()
    m.update(hash_str)
    return m.hexdigest()


def _get_es_actions(index, documents):
    """
    Create the Elasticsearch bulk actions from article dicts.
    If you provide a list to ID_SEQ_LIST, the hashes are copied there
    """
    for document in documents:
        for f in REQUIRED_FIELDS:
            if f not in document:
                raise ValueError("Field {f!r} not present in document {document}".format(**locals()))
        if '_id' not in document:
            document['_id'] = _get_hash(document)
        yield {
            "_index": index,
            **document
        }


def upload_documents(index: str, documents, columns: Mapping[str, str] = None) -> List[str]:
    """
    Upload documents to this index

    :param index: The name of the index (without prefix)
    :param documents: A sequence of article dictionaries
    :param columns: A mapping of field:type for column types
    :return: the list of document ids
    """
    if columns:
        mapping = {field: ES_MAPPINGS[type_] for (field, type_) in columns.items()}
        body = {"properties": mapping}
        es.indices.put_mapping(index=index, body=body)

    actions = list(_get_es_actions(index, documents))
    bulk(es, actions)
    return [action['_id'] for action in actions]


def get_document(index: str, doc_id: str, **kargs) -> dict:
    """
    Get a single document from this index

    :param index: The name of the index
    :param doc_id: The document id (hash)
    :return: the source dict of the document
    """
    return es.get(index=index, id=doc_id, **kargs)['_source']


def update_document(index: str, doc_id: str, fields: dict):
    """
    Update a single document


    :param index: The name of the index
    :param doc_id: The document id (hash)
    :param fields: a {field: value} mapping of fields to update
    """
    body = {"doc": fields}
    es.update(index=index, id=doc_id, body=body)


def get_fields(index: str) -> Mapping[str, str]:
    """
    Get the field types in use in this index
    :param index:
    :return: a dictionary of field: type
    """
    r = es.indices.get_mapping(index=index)
    fields = r[index]['mappings']['properties']
    return {k:v['type'] for (k,v) in fields.items()}


def field_type(index: str, field_name: str) -> str:
    """
    Get the field type for the given field.
    :return: a type name ('text', 'date', ..)
    """
    # TODO: [WvA] cache this as it should be invariant
    return get_fields(index)[field_name]


def get_values(index: str, field: str) -> List[str]:
    """
    Get the values for a given field (e.g. to populate list of filter values on keyword field)
    :param index: The index
    :param field: The field name
    :return: A list of values
    """
    body = {"size": 0, "aggs": {"values": {"terms": {"field": field}}}}
    r = es.search(index=index, body=body)
    return [x["key"] for x in r["aggregations"]["values"]["buckets"]]


def refresh(index: str):
    es.indices.refresh(index=index)


def index_exists(name: str) -> bool:
    """
    Check if an index with this name exists
    """
    return es.indices.exists(index=name)



"""
AmCAT4 REST API
"""

import logging
import sys
import io
import csv
import urllib.request

from amcat4 import auth
from amcat4.auth import Role, User
from amcat4.db import initialize_if_needed
from amcat4.elastic import setup_elastic, upload_documents
from amcat4.api import app
from amcat4.index import create_index, Index

SOTU_INDEX = "state_of_the_union"


def upload_test_data() -> Index:
    url = "https://raw.githubusercontent.com/ccs-amsterdam/example-text-data/master/sotu.csv"
    url_open = urllib.request.urlopen(url)
    csv.field_size_limit(sys.maxsize)
    csvfile = csv.DictReader(io.TextIOWrapper(url_open, encoding='utf-8'))

    # creates the index info on the sqlite db
    index = create_index(SOTU_INDEX)

    docs = [dict(title="{Year}: {President}".format(**row),
                 text=row['Text'],
                 date=row['Date'],
                 president=row['President'],
                 year=row['Year'],
                 party=row['Party'])
            for row in csvfile]
    columns = {"president": "keyword", "party": "keyword", "year": "int"}
    upload_documents(SOTU_INDEX, docs, columns)
    return index


import argparse

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument('--create-test-index', action='store_true')
args = parser.parse_args()

logging.basicConfig(format='[%(levelname)-7s:%(name)-15s] %(message)s', level=logging.INFO)
setup_elastic()
initialize_if_needed()
es_logger = logging.getLogger('elasticsearch')
es_logger.setLevel(logging.WARNING)
if not User.select().where(User.email == "admin").exists():
    logging.warning("**** No user detected, creating superuser admin:admin ****")
    auth.create_user("admin", "admin", Role.ADMIN)
if args.create_test_index:
    if not Index.select().where(Index.name == SOTU_INDEX):
        logging.info("**** Creating test index {} ****".format(SOTU_INDEX))
        admin = User.get(User.email == "admin")
        upload_test_data().set_role(admin, Role.ADMIN)
app.run(debug=True)



import logging

from peewee import *
import sys

# IF we're running nose tests, we want an in-memory db
if 'nose' in sys.modules.keys():
    logging.warning("I think you're unit testing: using in-memory db")
    db = SqliteDatabase(':memory:', pragmas={'foreign_keys': 1})
else:
    db = SqliteDatabase('amcat4.db', pragmas={'foreign_keys': 1})


def initialize_if_needed():
    from amcat4.auth import User
    from amcat4.index import Index, IndexRole
    db.create_tables([User, Index, IndexRole])



"""
Encapsulates elasticsearch indices  and provides authorisation through user roles

Authorisation rules:
- There are 5 increasing levels of authorisation: None, metareader, reader, writer, admin
- Users can have a global role and a role on any index. Every index can also have a guest role
- Global roles:
  - Writers can create new projects and users (with at most their own global role)
  - Admins can delete projects and assign themselves a role on any index role
- Index roles:
  - None means an index cannot be viewed or accessed in any way (the index is invisible to the user)
  - Metareader means the user can read all properties, do queries, etc., but cannot read the 'text' attribute
    (this is mostly intended to provide access to metadata fields of copyrighted material)
  - Reader can read all properties, do queries, etc., but cannot make changes
  - Writers can add/delete documents, add/delete users (up to their own level), and make other changes (but not delete)
  - Admins can do whatever they want, including deleting the index
- If a user does not have an explicit role on an index, the guest role (if any) is used
- An unauthorized user can still get guest roles, so it can see any indices with a guest role

Note that these rules are not enforced in this module, they should be enforced by the API!
"""
import logging

from peewee import Model, CharField, IntegerField, ForeignKeyField

from amcat4 import elastic
from amcat4.auth import Role, User
from amcat4.db import db


class Index(Model):
    name = CharField(unique=True)
    guest_role = IntegerField(null=True)

    class Meta:
        database = db

    def delete_index(self, delete_from_elastic=True):
        """
        Delete this index
        :param delete_from_elastic: if True (default), also delete the underlying elastic table
        """
        self.delete_instance()
        if delete_from_elastic:
            elastic._delete_index(self.name)
        indices = [i.name for i in self.select()]

    def has_role(self, user: User, role: Role) -> bool:
        """
        Checks whether the given User has at least the required Role
        """
        ir = IndexRole.get_or_none(IndexRole.user == user, IndexRole.index == self)
        actual_role = self.guest_role if ir is None else ir.role
        if not actual_role:
            return False
        else:
            return actual_role and actual_role >= role

    def set_role(self, user: User, role: Role) -> None:
        """
        Sets the role for the given new or existing user; set role to None to remove user from this index.
        This will create/update/delete the role as needed
        """
        ir = IndexRole.get_or_none(IndexRole.user == user, IndexRole.index == self)
        if ir:
            if role is None:
                ir.delete_instance()
            elif role != ir.role:
                ir.role = role
                ir.save()
        elif role is not None:
            IndexRole.create(user=user, index=self, role=role)


class IndexRole(Model):
    user = ForeignKeyField(User, on_delete="CASCADE")
    index = ForeignKeyField(Index, on_delete="CASCADE")
    role = IntegerField()

    class Meta:
        database = db
        indexes = (
            (('user', 'index'), True),  # unique constraint user & index
        )


def create_index(name: str, guest_role: Role = None, create_in_elastic=True, admin: User = None) -> Index:
    """
    Create a new index in both elastic and amcat4
    :param name: Name of the new index (in elastic and amcat4)
    :param guest_role: Guest role for this index, i.e. a user's role if no explicit role is assigned.
    :param admin: if given, add this User as admin
    :param create_in_elastic: if True (default), also create a new elastic index
    """
    if Index.select().where(Index.name == name).exists():
        raise ValueError("Index {name} is already registered".format(**locals()))
    if create_in_elastic:
        if elastic.index_exists(name):
            raise ValueError("Index {name} already exists in elastic".format(**locals()))
        elastic._create_index(name)
    elif not elastic.index_exists(name):
        raise ValueError("Index {name} does not exist in elastic".format(**locals()))

    index = Index.create(name=name, guest_role=guest_role)
    if admin:
        IndexRole.create(user=admin, index=index, role=Role.ADMIN)

    indices = [i.name for i in Index.select()]

    return index



"""
Provide authentication for AmCAT4 api

Authentication can work via password or token
See amcat4.index for authorisation rules
"""
import logging
from enum import IntEnum
from typing import Optional, Iterable, Mapping

import bcrypt
from itsdangerous import TimedJSONWebSignatureSerializer, SignatureExpired, BadSignature
from peewee import Model, CharField, IntegerField
from amcat4.db import db

SECRET_KEY = "NOT VERY SECRET YET!"


class Role(IntEnum):
    METAREADER = 10
    READER = 20
    WRITER = 30
    ADMIN = 40


class User(Model):
    email = CharField(unique=True)
    password = CharField()
    global_role = IntegerField(null=True)

    class Meta:
        database = db

    def create_token(self, expiration: int = None) -> str:
        """
        Create a new token for this user
        """
        s = TimedJSONWebSignatureSerializer(SECRET_KEY, expires_in=expiration)
        return s.dumps({'id': self.id})

    def has_role(self, role: Role) -> bool:
        """
        Check whether this user has at least the requested Role
        """
        return self.global_role and self.global_role >= role

    def indices(self, include_guest: bool = False) -> Mapping['Index', Role]:
        from amcat4.index import Index  # Prevent circular import
        indices = {i.index: Role(i.role) for i in self.indexrole_set.join(Index)}
        if include_guest:
            for i in Index.select().where(Index.guest_role != None):
                if i not in indices:
                    indices[i] = Role(i.guest_role)
        return indices

    @property
    def role(self):
        return self.global_role and Role(self.global_role)


def hash_password(password: str) -> str:
    return bcrypt.hashpw(password.encode("utf-8"), bcrypt.gensalt()).decode("utf-8")


def create_user(email: str, password: str, global_role: Role = None) -> User:
    """
    Create and return a new User with the given information
    """
    if global_role not in [None, Role.WRITER, Role.ADMIN]:
        raise ValueError("Global roles can only be None, Writer, or Admin")
    return User.create(email=email, password=hash_password(password), global_role=global_role)


def verify_user(email: str, password: str) -> Optional[User]:
    """
    Check that this user exists and can be authenticated with the given password, returning a User object
    :param email: Email address identifying the user
    :param password: Password to check
    :return: A User object if user could be authenticated, None otherwise
    """
    logging.info("Attempted login: {email}".format(**locals()))
    try:
        user = User.get(User.email == email)
    except User.DoesNotExist:
        logging.warning("User {email} not found!".format(**locals()))
        return None
    if bcrypt.checkpw(password.encode('utf-8'), user.password.encode("utf-8")):
        return user
    else:
        logging.warning("Incorrect password for user {email}".format(**locals()))


def verify_token(token: str) -> Optional[User]:
    """
    Check the token and return the authenticated user email
    :param token: The token to verify
    :return: a User object if user could be authenticated, None otherwise
    """
    s = TimedJSONWebSignatureSerializer(SECRET_KEY)
    try:
        result = s.loads(token)
    except (SignatureExpired, BadSignature):
        logging.exception("Token verification failed")
        return None
    logging.warning("TOKEN RESULT: {}" .format(result))
    return User.get(User.id == result['id'])



#!/usr/bin/env python

from distutils.core import setup

setup(
    name="amcat4",
    version="0.05",
    description="API for AmCAT4 Text Analysis",
    author="Wouter van Atteveldt",
    author_email="wouter@vanatteveldt.com",
    packages=["amcat4"],
    include_package_data=True,
    zip_safe=False,
    keywords=["API", "text"],
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Topic :: Text Processing",
    ],
    install_requires=[
        "Flask",
        "Flask-HTTPAuth",
        "flask-cors",
        "elasticsearch",
        "bcrypt",
        "peewee",
        "itsdangerous",
        "amcat4annotator>=0.10"
    ],
    extras_require={
        'dev': [
            'nose',
            'codecov',
        ]
    },
)



"""
Aggregate queries
"""
from collections import namedtuple
from datetime import datetime
from typing import Mapping, Iterable, Union

from amcat4.elastic import es, field_type
from amcat4.query import build_body


class _Axis:
    """
    Internal helper class that represents an aggregation axis
    """
    def __init__(self, index, field, interval=None):
        if isinstance(field, dict):
            self.field = field['field']
            self.interval = field.get('interval')
        else:
            self.field = field
            self.interval = interval
        self.ftype = field_type(index, self.field)

    def query(self):
        if self.interval:
            if self.ftype == "date":
                return {self.field: {"date_histogram": {"field": self.field, "interval": self.interval}}}
            else:
                return {self.field: {"histogram": {"field": self.field, "interval": self.interval}}}
        else:
            return {self.field: {"terms": {"field": self.field}}}

    def postprocess(self, value):
        if self.ftype == "date":
            value = datetime.utcfromtimestamp(value / 1000.)
            if self.interval in {"year", "month", "week", "day"}:
                value = value.date()
        return value


def _get_aggregates(index, sources, queries, filters, after_key=None):
    """
    Recursively get all buckets from a composite query
    """
    # [WvA] Not sure if we should get all results ourselves or expose the 'after' pagination.
    #       This might get us in trouble if someone e.g. aggregates on url or day for a large corpus
    after = {"after": after_key} if after_key else {}
    body = {"size": 0, "aggregations": {"aggr": {"composite": dict(sources=sources, **after)}}}
    if filters or queries:
        body['query'] = build_body(queries=queries, filters=filters)
    result = es.search(index=index, body=body)['aggregations']['aggr']
    yield from result['buckets']
    after_key = result.get('after_key')
    if after_key:
        yield from _get_aggregates(index, sources, queries, filters, after_key)


def query_aggregate(index: str, axis: Union[str, dict], *axes: Union[str, dict],
                    value="n", queries: Iterable[str] = None,
                    filters: Mapping[str, Mapping] = None) -> Iterable[namedtuple]:
    """
    Conduct an aggregate query.
    Note that interval queries also yield zero counts for intervening keys without value,
    but only if that is the last axis. [WvA] Not sure if this is desired

    :param index: The name of the elasticsearch index
    :param axis: The primary aggregation axis, should be the name of a field or a dict with keys 'field', 'interval'
    :param axes: Optional additional axes
    :param value: Name for the value 'column', default n. This might be expanded for other aggregate values (avg etc)
    :param queries: Optional query string
    :param filters: if not None, a dict of filters: {field: {'value': value}} or
                    {field: {'range': {'gte/gt/lte/lt': value, 'gte/gt/..': value, ..}}
    :return: a sequence of (axis-value, [axis2-value, ...], aggregate-value) tuples
    """
    axes = [_Axis(index, x) for x in (axis, ) + axes]
    names = [x.field for x in axes] + [value]
    nt = namedtuple("Bucket", names)
    sources = [axis.query() for axis in axes]
    for bucket in _get_aggregates(index, sources, queries, filters):
        result = [axis.postprocess(bucket['key'][axis.field]) for axis in axes]
        result += [bucket['doc_count']]
        yield nt(*result)
